﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using Windows.Foundation;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Input;
using System.Diagnostics;

namespace BasicDriveApp
{
    public class Joystick
    {
        public event EventHandler Stop;
        public event EventHandler<MovementEventArgs> Movement;

        private FrameworkElement m_puckControl;
        private TranslateTransform m_translateTransform;
        private int m_lastHeading = 0;
        private Point m_initialPoint;
        private long m_lastCommandSentTimeMs;

        public Joystick(FrameworkElement puck)
        {
            m_puckControl = puck;
            m_puckControl.PointerPressed += PointerPressed;
            m_puckControl.PointerMoved += PointerMoved;
            m_puckControl.PointerReleased += PointerReleased;
            m_puckControl.PointerCaptureLost += PointerReleased;
            m_puckControl.PointerCanceled += PointerReleased;

            m_translateTransform = new TranslateTransform();
            m_puckControl.RenderTransform = m_translateTransform;
        }

        private void PointerPressed(object sender, PointerRoutedEventArgs args)
        {
            Windows.UI.Input.PointerPoint pointer = args.GetCurrentPoint(null);
            if (pointer.Properties.IsLeftButtonPressed)
            {
                m_initialPoint = pointer.RawPosition;
                args.Handled = true;

                m_puckControl.CapturePointer(args.Pointer);
            }
        }

        private void PointerMoved(object sender, PointerRoutedEventArgs args)
        {
            Windows.UI.Input.PointerPoint pointer = args.GetCurrentPoint(null);
            if (pointer.Properties.IsLeftButtonPressed)
            {
                Point newPoint = pointer.RawPosition;
                Point delta = new Point(
                    newPoint.X - m_initialPoint.X,
                    newPoint.Y - m_initialPoint.Y);
                m_translateTransform.X = delta.X;
                m_translateTransform.Y = delta.Y;
                args.Handled = true;

                ConstrainToParent();

                SendCommand();
            }
        }

        private void ConstrainToParent()
        {
            FrameworkElement parent = m_puckControl.Parent as FrameworkElement;
            float radius = (float)Math.Min(parent.ActualWidth, parent.ActualHeight) / 2f;

            float distanceSq = (float)(
                m_translateTransform.X * m_translateTransform.X
                + m_translateTransform.Y * m_translateTransform.Y);

            float radiusSq = radius * radius;
            if (distanceSq > radiusSq)
            {
                float length = (float)Math.Sqrt(distanceSq);
                Point positionNorm = new Point(m_translateTransform.X / length, m_translateTransform.Y / length);

                m_translateTransform.X = positionNorm.X * radius;
                m_translateTransform.Y = positionNorm.Y * radius;
            }
        }

        private void PointerReleased(object sender, PointerRoutedEventArgs args)
        {
            m_translateTransform.X = 0;
            m_translateTransform.Y = 0;

            m_puckControl.ReleasePointerCapture(args.Pointer);

            SendStopCommand();

            args.Handled = true;
        }

        private void SendStopCommand()
        {
            Debug.WriteLine("Stop");

            if (Stop != null)
                Stop(this, EventArgs.Empty);
        }

        private void SendCommand()
        {
            float x = (float)m_translateTransform.X;
            float y = (float)m_translateTransform.Y;

            //Debug.WriteLine("{0}, {1}", x, y);

            double valueX = (x / 255);
            double valueY = (y / 255);

            //Debug.WriteLine("{0}, {1}\n", valueX, valueY);

            FrameworkElement parent = m_puckControl.Parent as FrameworkElement;
            Size size = new Size(parent.ActualWidth, parent.ActualHeight);

            x /= (float)size.Width * .5f;
            y /= (float)size.Height * .5f;

            float speed = x * x + y * y;
            speed = (speed == 0) ? 0 : (float)Math.Sqrt(speed);
            if (speed > 1f)
                speed = 1f;

            double rad = Math.Atan2((double)y, (double)x);
            rad += Math.PI / 2.0;
            double degrees = rad * 180.0 / Math.PI;
            int degreesCapped = (((int)degrees) + 360) % 360;

            // Send roll commands and limit to 10 Hz
            long milliseconds = DateTime.Now.Ticks / TimeSpan.TicksPerMillisecond;
            if ((milliseconds - m_lastCommandSentTimeMs) > 100)
            {
                if (Movement != null)
                    Movement(this, new MovementEventArgs(valueX, valueY));

                m_lastCommandSentTimeMs = milliseconds;
            }

            m_lastHeading = degreesCapped;
        }
    }

    public class MovementEventArgs : EventArgs
    {
        public double X { get; private set; }
        public double Y { get; private set; }

        public MovementEventArgs(double x, double y)
        {
            this.X = x;
            this.Y = y;
        }
    }
}
