﻿using WiFiDirectChat.Common;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;
using Windows.Networking.Proximity;
using Windows.Networking.Sockets;
using Windows.Storage.Streams;
using System.Threading.Tasks;
using Windows.UI.Popups;
using Windows.System;
using Windows.UI.Text;

namespace WiFiDirectChat
{
    public sealed partial class MainPage : Page
    {
        #region Properties

        private NavigationHelper navigationHelper;

        /// <summary>
        /// NavigationHelper is used on each page to aid in navigation and 
        /// process lifetime management
        /// </summary>
        public NavigationHelper NavigationHelper
        {
            get { return this.navigationHelper; }
        }

        #endregion

        public enum PairingContinuationAction
        {
            ContinueReading,
            CloseSocket
        }

        private PeerInformation peer;
        private StreamSocket socket;
        private DataReader reader;
        private DataWriter writer;

        private bool isConnected;

        public MainPage()
        {
            this.InitializeComponent();
            this.navigationHelper = new NavigationHelper(this);
            this.navigationHelper.LoadState += navigationHelper_LoadState;
        }

        protected async override void OnNavigatedTo(NavigationEventArgs e)
        {
            navigationHelper.OnNavigatedTo(e);

            if (PeerFinder.SupportedDiscoveryTypes.HasFlag(PeerDiscoveryTypes.Browse))
            {
                PeerFinder.Start();
                PeerFinder.ConnectionRequested += PeerFinder_ConnectionRequested;

                var peers = await PeerFinder.FindAllPeersAsync();

                if (peers.Count > 0)
                {
                    statusTextBox.Text = "Peer identificato, connessione in corso...";

                    // Se è stato trovato un peer, si collega ad esso.
                    await this.HandleConnectionAsync(peers[0]);
                }
                else
                {
                    statusTextBox.Text = "Nessun peer trovato.";
                }
            }
        }

        private async Task HandleConnectionAsync(PeerInformation connectingPeer)
        {
            try
            {
                peer = connectingPeer;

                socket = await PeerFinder.ConnectAsync(connectingPeer);

                reader = new DataReader(socket.InputStream);
                writer = new DataWriter(socket.OutputStream);

                isConnected = true;
                this.ListenAsync(reader);

                statusTextBox.Text = "Connessione effettuata.";
            }
            catch
            { }

            if (!isConnected)
            {
                var dialog = new MessageDialog("Errore durante la connessione");
                await dialog.ShowAsync();
            }
        }

        private async void ListenAsync(DataReader socketReader)
        {
            var readingSucceeded = false;

            try
            {
                // Controlla se lo stream contiene dati, verificando la presenza di un numero che contiene
                // la lunghezza del testo.
                var bytesRead = await socketReader.LoadAsync(sizeof(uint));
                if (bytesRead > 0)
                {
                    // Legge il numero di caratteri presenti nello stream e, subito dopo, carica dallo stream
                    // il numero di byte specificati.
                    var len = socketReader.ReadUInt32();
                    bytesRead = await socketReader.LoadAsync(len);
                    if (bytesRead > 0)
                    {
                        // Se effettivamente ci sono dati, li legge in formato stringa.
                        var message = socketReader.ReadString(len);

                        // Richiama il metodo che gestisce il messaggio ricevuto.
                        var response = this.Receive(message);
                        if (response == PairingContinuationAction.ContinueReading)
                        {
                            // Sono previsti ulteriori messaggi, quindi avvia una nuova lettura.
                            readingSucceeded = true;
                            this.ListenAsync(socketReader);
                        }
                    }
                }
            }
            catch
            { }

            if (!readingSucceeded)
                this.CloseSocket();
        }

        private PairingContinuationAction Receive(string message)
        {
            this.InsertPartnerMessage(message);
            return PairingContinuationAction.ContinueReading;
        }

        private async Task SendAsync(string message)
        {
            var sendingSucceeded = false;

            if (isConnected && !string.IsNullOrEmpty(message))
            {
                try
                {
                    // Calcola il numero di byte da inviare.
                    var len = writer.MeasureString(message);
                    writer.WriteUInt32(len);

                    // Invia il messaggio vero e proprio.
                    writer.WriteString(message);
                    var bytesWritten = await writer.StoreAsync();

                    sendingSucceeded = bytesWritten > 0;
                }
                catch
                { }
            }

            if (!sendingSucceeded)
                this.CloseSocket();
        }

        private async void PeerFinder_ConnectionRequested(object sender, ConnectionRequestedEventArgs args)
        {
            // E' arrivata la richiesta di connessione da un peer remoto.
            await Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.Normal, async () =>
            {
                statusTextBox.Text = "Richiesta di connessione da un peer. Connessione in corso...";

                await this.HandleConnectionAsync(args.PeerInformation);
            });
        }

        private void CloseSocket()
        {
            if (reader != null)
            {
                reader.DetachStream();
                reader.Dispose();
                reader = null;
            }

            if (writer != null)
            {
                writer.DetachStream();
                writer.Dispose();
                writer = null;
            }

            if (socket != null)
            {
                socket.Dispose();
                socket = null;
            }

            peer = null;
            isConnected = false;
        }

        #region Helpers 

        private void InsertPartnerMessage(string text)
        {
            var item = new ListBoxItem
            {
                HorizontalAlignment = HorizontalAlignment.Left,
                Margin = new Thickness(5, -5, 0, -5)
            };

            var panel = new StackPanel();
            panel.Children.Add(new TextBlock
            {
                Text = peer.DisplayName,
                FontWeight = FontWeights.SemiBold,
            });
            panel.Children.Add(new TextBlock { Text = text });

            item.Content = panel;
            messagesListBox.Items.Add(item);
            this.ScrollToBottom();
        }

        private void InsertMyMessage(string text)
        {
            var item = new ListBoxItem
            {
                HorizontalAlignment = HorizontalAlignment.Right,
                Margin = new Thickness(0, -5, 20, -5)
            };

            var panel = new StackPanel();
            panel.Children.Add(new TextBlock
            {
                Text = PeerFinder.DisplayName,
                FontWeight = FontWeights.SemiBold
            });
            panel.Children.Add(new TextBlock { Text = text });

            item.Content = panel;
            messagesListBox.Items.Add(item);
            this.ScrollToBottom();
        }

        private void ScrollToBottom()
        {
            var selectedIndex = messagesListBox.Items.Count - 1;
            if (selectedIndex < 0)
                return;

            messagesListBox.SelectedIndex = selectedIndex;
            messagesListBox.UpdateLayout();

            messagesListBox.ScrollIntoView(messagesListBox.SelectedItem);
            messagesListBox.SelectedIndex = -1;
        }

        #endregion

        #region NavigationHelper registration

        /// <summary>
        /// Populates the page with content passed during navigation.  Any saved state is also
        /// provided when recreating a page from a prior session.
        /// </summary>
        /// <param name="sender">
        /// The source of the event; typically <see cref="NavigationHelper"/>
        /// </param>
        /// <param name="e">Event data that provides both the navigation parameter passed to
        /// <see cref="Frame.Navigate(Type, Object)"/> when this page was initially requested and
        /// a dictionary of state preserved by this page during an earlier
        /// session.  The state will be null the first time a page is visited.</param>
        private void navigationHelper_LoadState(object sender, LoadStateEventArgs e)
        {

        }

        /// The methods provided in this section are simply used to allow
        /// NavigationHelper to respond to the page's navigation methods.
        /// 
        /// Page specific logic should be placed in event handlers for the  
        /// <see cref="GridCS.Common.NavigationHelper.LoadState"/>
        /// and <see cref="GridCS.Common.NavigationHelper.SaveState"/>.
        /// The navigation parameter is available in the LoadState method 
        /// in addition to page state preserved during an earlier session.

        protected override void OnNavigatedFrom(NavigationEventArgs e)
        {
            navigationHelper.OnNavigatedFrom(e);
        }

        #endregion

        private async void sendButton_Click(object sender, RoutedEventArgs e)
        {
            if (!string.IsNullOrWhiteSpace(messageTextBox.Text))
            {
                await this.SendAsync(messageTextBox.Text);
                this.InsertMyMessage(messageTextBox.Text);
                messageTextBox.Text = string.Empty;
            }
        }

        private void messageTextBox_KeyUp(object sender, KeyRoutedEventArgs e)
        {
            if (e.Key == VirtualKey.Enter)
                sendButton_Click(sender, new RoutedEventArgs());
        }
    }
}